from contextlib import contextmanager
from typing import Generator

from openai import APIError

from .typer import typer_raise

__all__ = [
    'openai_handler',
]


@contextmanager
def openai_handler() -> Generator[None, None, None]:
    try:
        yield
    except APIError as e:
        typer_raise(f'OpenAI API error:\n{e}')



from __future__ import annotations
from dataclasses import dataclass
from typing import Callable, cast

from rich.console import Console
from rich.text import Text

from reling.tts import TTSVoiceClient
from reling.types import Reader, Speed
from reling.utils.console import clear_current_line, input_and_erase
from reling.utils.prompts import ENTER, Prompt, PromptOption
from reling.utils.values import coalesce, ensure_not_none
from .colors import fade

__all__ = [
    'output',
    'SentenceData',
]

NA = fade('N/A')

PLAY_PROMPT_TITLE = 'Play'
NORMAL_SPEED = 'normal speed'
SLOWLY = 'slowly'
REPLAY = 'replay'

INTERRUPTED = f'INTERRUPTED ({ENTER} to continue)'


@dataclass
class SentenceData:
    print_text: str | Text | None = None
    print_prefix: str | Text | None = None
    reader: Reader | None = None
    reader_id: str | None = None

    @staticmethod
    def from_tts(
            text: str | None,
            client: TTSVoiceClient | None,
            *,
            print_text: str | Text | None = None,
            print_prefix: str | Text | None = None,
            reader_id: str | None = None,
    ) -> SentenceData:
        return SentenceData(
            print_text=coalesce(print_text, text),
            print_prefix=print_prefix,
            reader=client.get_reader(text) if client and (text is not None) else None,
            reader_id=reader_id,
        )


@dataclass
class ReaderWithSpeed:
    reader: Reader
    speed: Speed


@dataclass
class ReaderWithId:
    reader: Reader
    id: str


type OutputPromptOption = ReaderWithSpeed | Callable[[], None]
type OutputPrompt = Prompt[OutputPromptOption]


def add_single_sentence_options(prompt: OutputPrompt, reader: Reader) -> None:
    """Attach the options for a single sentence to the prompt: '[n]ormal speed | [s]lowly'."""
    prompt.add_option(PromptOption(
        description=NORMAL_SPEED,
        action=ReaderWithSpeed(reader, Speed.NORMAL),
    ))
    prompt.add_option(PromptOption(
        description=SLOWLY,
        action=ReaderWithSpeed(reader, Speed.SLOW),
    ))


def add_multi_sentence_options(prompt: OutputPrompt, readers: list[ReaderWithId]) -> None:
    """Attach the options for multiple sentences to the prompt: '[i]mproved | [is] | [o]riginal | [os]'."""
    for reader in readers:
        prompt.add_option(PromptOption(
            description=reader.id,
            action=ReaderWithSpeed(reader.reader, Speed.NORMAL),
            modifiers={
                SLOWLY: ReaderWithSpeed(reader.reader, Speed.SLOW),
            },
        ))


def add_extra_options(prompt: OutputPrompt, extra_options: list[PromptOption[Callable[[], None]]]) -> None:
    """Attach the extra options to the prompt."""
    for option in extra_options:
        prompt.add_option(option)


def construct_play_prompt(
        sentences_with_readers: list[SentenceData],
        current: ReaderWithSpeed | None,
        multi_sentence: bool,
        extra_options: list[PromptOption[Callable[[], None]]],
) -> OutputPrompt:
    """
    Construct a prompt for the user to choose the next sentence to read and the speed of the reading.
    :raises ValueError: If reader_id is not provided for a sentence with a reader in a multi-sentence output.
    """
    prompt: OutputPrompt = Prompt(PLAY_PROMPT_TITLE)
    if multi_sentence:
        add_multi_sentence_options(prompt, [
            ReaderWithId(
                reader=ensure_not_none(sentence.reader),
                id=ensure_not_none(sentence.reader_id),
            )
            for sentence in sentences_with_readers
        ])
    else:
        add_single_sentence_options(prompt, sentences_with_readers[0].reader)
    if current:
        prompt.add_option(PromptOption(
            description=REPLAY,
            action=current,
        ))
    add_extra_options(prompt, extra_options)
    return prompt


def do_prompt(prompt: OutputPrompt) -> OutputPromptOption | None:
    """Prompt the user and return the chosen option."""
    while True:
        try:
            return prompt.prompt()
        except KeyboardInterrupt:
            input_and_erase(INTERRUPTED)


def invoke(option: OutputPromptOption) -> None:
    """Invoke the reader or the function."""
    try:
        if isinstance(option, ReaderWithSpeed):
            option.reader(option.speed)
        else:
            option()
    except KeyboardInterrupt:
        pass
    clear_current_line()  # Otherwise the input made during the operation will get displayed twice


def output(*sentences: SentenceData, extra_options: list[PromptOption[Callable[[], None]]] | None = None) -> None:
    """
    Output the sentences, reading them if a reader is provided.
    If multiple readers are provided, the user can choose which sentence to read next.
    The user can also choose the speed of the reading.

    :raises ValueError: If reader_id is not provided for a sentence with a reader in a multi-sentence output.
    """
    extra_options = extra_options or []
    console = Console(highlight=False)
    for sentence in sentences:
        console.print(sentence.print_prefix or '', end='')
        console.print(coalesce(sentence.print_text, cast(str | Text, NA)))
    multi_sentence = len(sentences) > 1
    if sentences_with_readers := [sentence for sentence in sentences if sentence.reader]:
        current = ReaderWithSpeed(sentences_with_readers[0].reader, Speed.NORMAL) if len(sentences) == 1 else None
        while True:
            if current:
                invoke(current)
            current = do_prompt(construct_play_prompt(sentences_with_readers, current, multi_sentence, extra_options))
            if not current:
                break
    elif multi_sentence or extra_options:
        prompt: OutputPrompt = Prompt()
        add_extra_options(prompt, extra_options)
        while True:
            if current := do_prompt(prompt):
                invoke(current)
            else:
                break



from contextlib import contextmanager
from functools import partial
from pathlib import Path
from typing import Generator, Mapping
import wave

from reling.types import Speed
from .pyaudio import get_audio, get_stream

__all__ = [
    'FILE_EXTENSION',
    'play',
    'record',
]

FILE_EXTENSION = '.wav'

CHUNK_SIZE = 1024

RECORD_CHANNELS = 1
RECORD_RATE = 16000
RECORD_BITS_PER_SAMPLE = 16


@contextmanager
def write(
        file: Path,
        channels: int,
        sample_width: int,
        rate: float,
) -> Generator[wave.Wave_write, None, None]:
    """Initialize a WAV file for writing, yield it, and close it afterward."""
    wav = wave.open(str(file.absolute()), 'wb')
    try:
        wav.setnchannels(channels)
        wav.setsampwidth(sample_width)
        wav.setframerate(rate)
        yield wav
    finally:
        wav.close()


def play(file: Path, speed: Speed = Speed.NORMAL) -> None:
    """Play a WAV file."""
    with (
        wave.open(str(file.absolute()), 'rb') as wav,
        get_audio() as pyaudio,
        get_stream(
            pyaudio=pyaudio,
            format=pyaudio.audio.get_format_from_width(wav.getsampwidth()),
            channels=wav.getnchannels(),
            rate=int(wav.getframerate() * speed.value),
            output=True,
        ) as stream,
    ):
        data = wav.readframes(CHUNK_SIZE)
        while data:
            stream.write(data)
            data = wav.readframes(CHUNK_SIZE)


def record_on_data(
        wav: wave.Wave_write,
        flag_to_return: int,
        data: bytes,
        _frame_count: int,
        _time_info: Mapping[str, float],
        _status_flags: int,
) -> tuple[None, int]:
    """Implement the stream callback for recording audio."""
    wav.writeframes(data)
    return None, flag_to_return


@contextmanager
def record(file: Path) -> Generator[None, None, None]:
    """Record audio in real time and save it to a file."""
    with (
        write(
            file,
            channels=RECORD_CHANNELS,
            sample_width=RECORD_BITS_PER_SAMPLE // 8,
            rate=RECORD_RATE,
        ) as wav,
        get_audio() as pyaudio,
        get_stream(
            pyaudio=pyaudio,
            format=pyaudio.paInt16,
            channels=RECORD_CHANNELS,
            rate=RECORD_RATE,
            input=True,
            frames_per_buffer=CHUNK_SIZE,
            stream_callback=partial(record_on_data, wav, pyaudio.paContinue),
        ),
    ):
        yield



from pathlib import Path

from sqlalchemy.sql import exists, func

from reling.db import single_session
from reling.db.models import Language
from reling.utils.csv import read_csv
from reling.utils.strings import replace_prefix_casing

__all__ = [
    'find_language',
    'find_languages_by_prefix',
    'populate_languages',
]


def populate_languages(data: Path) -> None:
    """Populate the languages table with data from the CSV file, if the table is empty."""
    with single_session() as session:
        if not session.query(exists().where(Language.id.is_not(None))).scalar():
            for language in read_csv(
                data,
                ['id', 'short_code', 'name', 'extra_name_a', 'extra_name_b'],
                empty_as_none=True,
            ):
                session.add(Language(**language))
        session.commit()


def find_language(language: str) -> Language | None:
    """Find a language, either by its ID, short code, or name (case-insensitive), applying filters sequentially."""
    lower = func.lower(language)
    conditions = [
        Language.id == language,
        Language.short_code == language,
        func.lower(Language.name) == lower,
        func.lower(Language.extra_name_a) == lower,
        func.lower(Language.extra_name_b) == lower,
    ]

    with single_session() as session:
        for condition in conditions:
            if result := session.query(Language).where(condition).first():
                return result

    return None


def find_languages_by_prefix(prefix: str) -> list[str]:
    """Find language IDs, short codes, and names that start with the given prefix."""
    lower = func.lower(prefix)
    with single_session() as session:
        return sorted([
            language.id
            for language in session.query(Language).where(Language.id.startswith(prefix)).all()
        ] + [
            language.short_code
            for language in session.query(Language).where(Language.short_code.startswith(prefix)).all()
        ] + [
            replace_prefix_casing(language.name, prefix)
            for language in session.query(Language).where(func.lower(Language.name).startswith(lower)).all()
        ] + [
            replace_prefix_casing(language.extra_name_a, prefix)
            for language in session.query(Language).where(func.lower(Language.extra_name_a).startswith(lower)).all()
        ] + [
            replace_prefix_casing(language.extra_name_b, prefix)
            for language in session.query(Language).where(func.lower(Language.extra_name_b).startswith(lower)).all()
        ])



from __future__ import annotations
from enum import StrEnum

__all__ = [
    'ContentCategory',
    'Gender',
    'Level',
]


class ContentCategory(StrEnum):
    TEXT = 'text'
    DIALOGUE = 'dialogue'


class Level(StrEnum):
    BASIC = 'basic'
    INTERMEDIATE = 'intermediate'
    ADVANCED = 'advanced'

    def get_included_levels(self) -> list[Level]:
        match self:
            case Level.BASIC:
                return [Level.BASIC]
            case Level.INTERMEDIATE:
                return [Level.BASIC, Level.INTERMEDIATE]
            case Level.ADVANCED:
                return [Level.BASIC, Level.INTERMEDIATE, Level.ADVANCED]
            case _:
                raise NotImplementedError


class Gender(StrEnum):
    MALE = 'male'
    FEMALE = 'female'
    NONBINARY = 'nonbinary'

    def describe(self) -> str:
        match self:
            case Gender.MALE:
                return 'a male'
            case Gender.FEMALE:
                return 'a female'
            case Gender.NONBINARY:
                return 'a nonbinary person'
            case _:
                raise NotImplementedError



from typing import Generator

from lcs2 import lcs_indices
from tqdm import tqdm

from reling.app.exceptions import AlgorithmException
from reling.config import MAX_SCORE
from reling.db.enums import ContentCategory
from reling.db.models import Language
from reling.gpt import GPTClient
from reling.helpers.scoring import calculate_diff_score
from reling.types import DialogueExchangeData, Promise
from reling.utils.english import pluralize
from reling.utils.iterables import extract_items, group_items, intersperse, strict_zip
from reling.utils.transformers import add_numbering, apply, get_number, omit_empty, remove_numbering, strip
from reling.utils.values import wrap_in_list
from .types import ExchangeWithTranslation, PreScoreWithSuggestion, ScoreWithSuggestion, SentenceWithTranslation

__all__ = [
    'score_translations',
]

NA = 'N/A'


def build_prompt_translation(
        category: ContentCategory,
        source_language: Language,
        target_language: Language,
        blocks: list[str],
        translations: list[str | None],
) -> str:
    """Build a prompt for scoring translations."""
    # Speaker turns in dialogues are "graded" as well so that the model appreciates the context.
    total = len(blocks)
    numbers = [get_number(index) for index, translation in enumerate(translations) if translation is not None]
    n = len(numbers)
    return '\n'.join([
        f'Below {'is' if total == 1 else 'are'} {total} {pluralize('sentence', total)} from a {category.value} '
        f'in {source_language.name}, along with {'a ' if n == 1 else ''}{pluralize('translation', n)} '
        f'of {n} of them into {target_language.name} made by a language learner.',

        f'Score {'the' if n == 1 else 'each'} translation on a scale from 0 to {MAX_SCORE}. '
        f'If {'the' if n == 1 else 'a'} translation is empty, very short, or poor, assign a low score. ',
        f'If {'the' if n == 1 else 'a'} translation is less than perfect, suggest a minimally modified version that '
        f'would deserve a {MAX_SCORE}.',

        f'{'Provide' if n == 1 else 'For each translation, provide'} your feedback on exactly five lines '
        f'(without adding bullet points or dashes in front of them):',
        f'- number of the sentence being evaluated on the first line (one of {', '.join(numbers)});',
        f'- original sentence on the second line;',  # The first three lines help improve the model's performance
        f'- learner\'s translation on the third line;',
        f'- score (just the number) on the fourth line;',
        f'- suggested modified translation (or "{NA}") on the fifth line.',

        *([f'Provide this feedback for each of the {n} translations, on exactly {n * 5} lines.'] if n > 1 else []),
        f'Say nothing else.',
        f'',
        f'The original {category.value} is:',
        *apply(add_numbering, blocks),
        f'',
        f'The translations are:',
        *(add_numbering(translation, index)
          for index, translation in enumerate(translations)
          if translation is not None),
    ])


def parse_scoring(string_score: str, suggestion: str) -> PreScoreWithSuggestion:
    """
    Parse the score and suggestion from the model output.
    :raises AlgorithmException: If there is an issue with the output of the model.
    """
    try:
        score = int(string_score)
    except ValueError:
        raise AlgorithmException(f'Could not parse the score as an integer from the model output: {string_score}.')
    if score < 0 or score > MAX_SCORE:
        raise AlgorithmException(f'The score {score} given by the model is not in the range from 0 to {MAX_SCORE}.')
    return PreScoreWithSuggestion(
        score=score,
        suggestion=(suggestion or None) if suggestion != NA else None,
    )


def ask_and_parse_translation(gpt: GPTClient, prompt: str) -> Generator[PreScoreWithSuggestion, None, None]:
    """
    Ask the model to score translations and parse the output.
    :raises AlgorithmException: If there is an issue with the output of the model.
    """
    for _, _, _, string_score, suggestion in group_items(gpt.ask(
        prompt,
        creative=False,
        transformers=[strip, omit_empty, remove_numbering],
    ), 5):
        yield parse_scoring(string_score, suggestion)


def build_prompt_averaging(language: Language, sentence: str, a: str, b: str) -> str:
    """Build a prompt for scoring an "averaged" translation."""
    return '\n'.join([
        f'Below are two nearly identical sentences in {language.name}:',

        add_numbering(a, 0),
        add_numbering(b, 1),

        f'A learner of {language.name} briefly viewed both sentences and was then asked to reproduce a similar '
        f'sentence from memory. The learner\'s response is provided below:',

        f'"""{sentence}"""',

        f'Score the learner\'s response on a scale from 0 to {MAX_SCORE}. Deduct points if the response contains '
        f'grammatical errors or does not convey the same meaning as the original sentences.',

        f'If the response is less than perfect, suggest a minimally modified version of it that would deserve a score '
        f'of {MAX_SCORE}.',

        f'Provide your feedback on exactly two lines (without adding bullet points or dashes in front of them):',
        f'- the score (just the number) on the first line;',
        f'- the suggested improved response (or "{NA}") on the second line (do not enclose it in quotes).',
    ])


def ask_and_parse_averaging(gpt: GPTClient, prompt: str) -> PreScoreWithSuggestion:
    """
    Ask the model to score an "averaged" translation and parse the output.
    :raises AlgorithmException: If there is an issue with the output of the model.
    """
    for string_score, suggestion in group_items(gpt.ask(
            prompt,
            creative=False,
            transformers=[strip, omit_empty, remove_numbering],
    ), 2):
        return parse_scoring(string_score, suggestion)
    raise AlgorithmException('The model did not provide a response.')


def lcs_indices_a(a: str, b: str) -> set[int | tuple[int, int]]:
    """
    Return a set of indices and consecutive index pairs in `a` that are part of the longest common subsequence with `b`.
    Consecutive index pairs are included only if the corresponding indices in `b` are also consecutive.
    """
    result: set[int | tuple[int, int]] = set()
    last_a_index = last_b_index = None
    for a_index, b_index in lcs_indices(a, b):
        result.add(a_index)
        if last_a_index == a_index - 1 and last_b_index == b_index - 1:
            result.add((last_a_index, a_index))
        last_a_index, last_b_index = a_index, b_index
    return result


def finalize_scoring(
        provided_translation: str,
        default_score: int | None = None,
        default_suggestion: str | None | None = None,
        perfect_options: set[str] | None = None,
) -> ScoreWithSuggestion:
    """
    Finalize the scoring by comparing the provided translation with the default suggestion and the perfect options
    and returning the best score and suggestion.
    """
    if default_score is None and default_suggestion is None and perfect_options is None:
        raise ValueError('At least one of the optional arguments must be provided.')

    if not provided_translation:
        return ScoreWithSuggestion(0, None)

    options = wrap_in_list(default_suggestion) + sorted(perfect_options - set(wrap_in_list(default_suggestion)))
    option_scores = [calculate_diff_score(provided_translation, option) for option in options]
    best_option_index = option_scores.index(max(option_scores)) if option_scores else None

    score, suggestion = (
        (option_scores[best_option_index], options[best_option_index])
        if option_scores and (default_score is None or option_scores[best_option_index] >= default_score)
        else (default_score, default_suggestion)
    )

    return ScoreWithSuggestion(
        score=score,
        suggestion=suggestion if suggestion != provided_translation and score > 0 else None,
    )


def fix_scoring(
        gpt: GPTClient,
        language: Language,
        provided_translation: str,
        original_translation: str,
        perfect_options: set[str],
        score: PreScoreWithSuggestion,
) -> ScoreWithSuggestion:
    """
    Fix the scoring by comparing the provided translation with the original translation and the suggested translation,
    as well as the perfect options, and returning the best score and suggestion.
    """
    default = (
        score if (score.suggestion is None
                  # If the provided translation shares as much or more common characters (individual indices)
                  # and omissions (pairs of consecutive indices) with the suggestion as with the original translation,
                  # proceed with the current score; otherwise, recalculate the score using "averaging":
                  or lcs_indices_a(provided_translation, score.suggestion)
                  >= lcs_indices_a(provided_translation, original_translation))
        else ask_and_parse_averaging(
            gpt,
            build_prompt_averaging(language, provided_translation, a=original_translation, b=score.suggestion),
        )
    )
    return finalize_scoring(
        provided_translation,
        default.score,
        default.suggestion,
        perfect_options,
    )


def extract_translation(translation: str | DialogueExchangeData) -> str:
    """Extract the translation from a sentence or a user turn in a dialogue."""
    return translation.user if isinstance(translation, DialogueExchangeData) else translation


def map_to_null_if_not[T](condition: bool, items: list[T]) -> list[T | None]:
    """Map the items to `None` if the condition is not met."""
    return [item if condition else None for item in items]


def score_offline(
        items: list[SentenceWithTranslation] | list[ExchangeWithTranslation],
        original_translations: list[str] | list[DialogueExchangeData],
        previous_perfect: list[set[str]],
) -> Generator[ScoreWithSuggestion | None, None, None]:
    """Score the translations of a text or user turns in a dialogue offline."""
    for item, original_translation, perfect in zip(items, original_translations, previous_perfect):
        yield finalize_scoring(
            provided_translation=item.input.text,
            default_score=None,
            default_suggestion=extract_translation(original_translation),
            perfect_options=perfect,
        ) if item.input else None


def score_with_gpt(
        category: ContentCategory,
        gpt: Promise[GPTClient],
        items: list[SentenceWithTranslation] | list[ExchangeWithTranslation],
        original_translations: list[str] | list[DialogueExchangeData],
        previous_perfect: list[set[str]],
        source_language: Language,
        target_language: Language,
) -> Generator[ScoreWithSuggestion | None, None, None]:
    """
    Score the translations of a text or user turns in a dialogue with the help of a GPT model.
    :raises AlgorithmException: If there is an issue with the output of the model.
    """
    translated_indices_set = {index for index in range(len(items)) if items[index].input}
    empty_translation_indices_set = {
        index for index in range(len(items))
        if index in translated_indices_set and not items[index].input.text
    }
    indices = [
        index for index in range(len(items))
        if index in translated_indices_set and index not in empty_translation_indices_set
        and items[index].input.text not in {extract_translation(original_translations[index])} | previous_perfect[index]
    ]
    indices_set = set(indices)
    client, prompt = (gpt(), build_prompt_translation(
        category=category,
        source_language=source_language,
        target_language=target_language,
        blocks=[block for item in items for block in item.all()],
        translations=[translation for index, item in enumerate(items)
                      for translation in map_to_null_if_not(index in indices_set, item.input_within_all())],
    )) if indices_set else (None, None)
    outer: list[
        tuple[SentenceWithTranslation, str, set[str], PreScoreWithSuggestion] |
        tuple[ExchangeWithTranslation, DialogueExchangeData, set[str], PreScoreWithSuggestion] |
        None
    ] = [None] * (len(items) - len(indices))
    for index, data in enumerate(intersperse(outer, zip(strict_zip(
            AlgorithmException('The model returned an unexpected number of results.'),
            extract_items(items, indices),
            extract_items(original_translations, indices),
            extract_items(previous_perfect, indices),
            tqdm(
                ask_and_parse_translation(client, prompt),
                desc='Scoring translations',
                total=len(indices),
                leave=False,
            ) if client and prompt else [],
    ), indices))):
        if data is None:
            yield (ScoreWithSuggestion(0 if index in empty_translation_indices_set else MAX_SCORE, None)
                   if index in translated_indices_set else None)
        else:
            item, original_translation, perfect, pre_score = data
            assert client is not None
            yield fix_scoring(
                client,
                target_language,
                item.input.text,
                extract_translation(original_translation),
                perfect,
                pre_score,
            )


def score_translations(
        category: ContentCategory,
        gpt: Promise[GPTClient],
        items: list[SentenceWithTranslation] | list[ExchangeWithTranslation],
        original_translations: list[str] | list[DialogueExchangeData],
        previous_perfect: list[set[str]],
        source_language: Language,
        target_language: Language,
        offline: bool,
) -> Generator[ScoreWithSuggestion | None, None, None]:
    """
    Score the translations of a text or user turns in a dialogue and provide suggestions for improvement.
    `None`s are yielded for the sentences or turns that are not translated.
    :raises AlgorithmException: If there is an issue with the scoring algorithm.
    """
    if offline:
        yield from score_offline(
            items=items,
            original_translations=original_translations,
            previous_perfect=previous_perfect,
        )
    else:
        yield from score_with_gpt(
            category=category,
            gpt=gpt,
            items=items,
            original_translations=original_translations,
            previous_perfect=previous_perfect,
            source_language=source_language,
            target_language=target_language,
        )



from reling.db.models import Dialogue, Language, Text
from .streaks import compute_pair_streaks

__all__ = [
    'get_skipped_indices',
]


def get_skipped_indices(
        content: Text | Dialogue,
        source_language: Language,
        target_language: Language,
        skip_after: int,
) -> set[int]:
    """Get the indices of sentences that should be skipped for a specified language pair."""
    streaks = compute_pair_streaks(content, source_language, target_language)
    return {index for index, streak in enumerate(streaks) if streak.count >= skip_after}



from datetime import datetime, timedelta, UTC

import pytz

__all__ = [
    'DATE_FORMAT',
    'format_time',
    'format_time_delta',
    'local_to_utc',
    'now',
    'TIME_FORMAT',
]

DATE_FORMAT = '%Y-%m-%d'
TIME_FORMAT = f'{DATE_FORMAT}, %H:%M'


def now() -> datetime:
    return datetime.now(UTC).replace(tzinfo=None)


def local_to_utc(time: datetime) -> datetime:
    return time.astimezone(UTC).replace(tzinfo=None)


def format_time(time: datetime, omit_zero_time: bool = False) -> str:
    localized = pytz.utc.localize(time).astimezone()
    return localized.strftime(
        DATE_FORMAT if omit_zero_time and localized.time() == datetime.min.time() else TIME_FORMAT,
    )


def format_time_delta(delta: timedelta) -> str:
    """Format timedelta as follows: '[[H:]MM:]SS'."""
    total_seconds = int(delta.total_seconds())

    hours, remainder = divmod(total_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    if hours > 0:
        return f'{hours}:{minutes:02}:{seconds:02}'
    else:
        return f'{minutes}:{seconds:02}'



from datetime import datetime
from typing import cast

from reling.app.app import app
from reling.app.default_content import set_default_content
from reling.app.types import ANSWERS_OPT, CONTENT_ARG, LANGUAGE_OPT, LANGUAGE_OPT_FROM
from reling.db.models import DialogueExam, DialogueExamResult, Language, TextExam, TextExamResult
from reling.helpers.colors import fade
from reling.helpers.scoring import format_average_score
from reling.helpers.typer import typer_raise
from reling.utils.tables import build_table, print_table, Table
from reling.utils.time import format_time, format_time_delta
from reling.utils.transformers import get_number

__all__ = [
    'history',
]

FROM = 'From'
TO = 'To'
TAKEN_AT = 'Taken at'
DURATION = 'Duration'
SENTENCES = 'Sents'
SCORE = 'Score'

SENTENCE_NUM = '#'
ANSWER = 'Answer'
EMPTY_ANSWER = fade('N/A')

AUDIO_OUTPUT = '🔈'
AUDIO_INPUT = '🎤'
IMAGE_INPUT = '📷'


def match(exam: TextExam | DialogueExam, from_: Language | None, to: Language | None) -> bool:
    return ((from_ is None or cast(Language, exam.source_language).id == from_.id)
            and (to is None or cast(Language, exam.target_language).id == to.id))


def get_sort_key(exam: TextExam | DialogueExam) -> tuple:
    return (
        cast(Language, exam.source_language).name,
        cast(Language, exam.target_language).name,
    )


def get_taken_at(exam: TextExam | DialogueExam) -> str:
    return format_time(cast(datetime, exam.started_at))


def get_default_table(exams: list[TextExam] | list[DialogueExam]) -> Table:
    """Generate a summary table for exam history."""
    return build_table(
        headers=[
            FROM,
            TO,
            TAKEN_AT,
            DURATION,
            SENTENCES,
            SCORE,
        ],
        justify={
            FROM: 'left',
            TO: 'left',
            TAKEN_AT: 'left',
            DURATION: 'right',
            SENTENCES: 'right',
            SCORE: 'right',
        },
        data=[{
            FROM: ' '.join([
                cast(Language, exam.source_language).name,
                *([AUDIO_OUTPUT] if exam.read_source else []),
            ]),
            TO: ' '.join([
                cast(Language, exam.target_language).name,
                *([AUDIO_OUTPUT] if exam.read_target else []),
                *([AUDIO_INPUT] if exam.listened else []),
                *([IMAGE_INPUT] if exam.scanned else []),
            ]),
            TAKEN_AT: get_taken_at(exam),
            DURATION: format_time_delta(exam.duration),
            SENTENCES: str(len(cast(list, exam.results))),
            SCORE: format_average_score(exam),
        } for exam in exams],
        group_by=[
            FROM,
            TO,
        ],
    )


def get_answers_table(exams: list[TextExam] | list[DialogueExam], include_from: bool, include_to: bool) -> Table:
    """Generate a detailed table showing answers for each exam."""
    if not exams:
        raise ValueError('No exams provided.')
    size = exams[0].content.size
    by_langs_and_index: dict[
        tuple[Language, Language],
        list[list[tuple[TextExamResult | DialogueExamResult, TextExam | DialogueExam]]],
    ] = {}
    for exam in exams:
        for result in exam.results:
            by_langs_and_index.setdefault(
                (exam.source_language, exam.target_language),
                [[] for _ in range(size)],
            )[result.index].append((result, exam))
    return build_table(
        headers=[
            *([FROM] if include_from else []),
            *([TO] if include_to else []),
            SENTENCE_NUM,
            TAKEN_AT,
            SCORE,
            ANSWER,
        ],
        justify={
            **({FROM: 'left'} if include_from else {}),
            **({TO: 'left'} if include_to else {}),
            SENTENCE_NUM: 'right',
            TAKEN_AT: 'left',
            SCORE: 'right',
            ANSWER: 'left',
        },
        data=[
            {
                **({FROM: source_language.name} if include_from else {}),
                **({TO: target_language.name} if include_to else {}),
                SENTENCE_NUM: get_number(sentence_index),
                TAKEN_AT: get_taken_at(exam),
                SCORE: str(result.score),
                ANSWER: result.answer or EMPTY_ANSWER,
            }
            for (source_language, target_language), items in by_langs_and_index.items()
            for sentence_index, results in enumerate(items)
            for result, exam in results
        ],
        group_by=[
            *([FROM] if include_from else []),
            *([TO] if include_to else []),
            SENTENCE_NUM,
        ],
    )


@app.command()
def history(
        content: CONTENT_ARG,
        from_: LANGUAGE_OPT_FROM = None,
        to: LANGUAGE_OPT = None,
        answers: ANSWERS_OPT = False,
) -> None:
    """Display exam history, optionally filtered by source or target language."""
    set_default_content(content)
    exams = sorted(
        filter(lambda e: match(e, from_, to), content.exams),
        key=get_sort_key,
    )
    if exams:
        table = get_answers_table(
            exams,
            include_from=from_ is None,
            include_to=to is None,
        ) if answers else get_default_table(exams)
        print_table(table)
    else:
        typer_raise('No exams found.', is_error=False)



from contextlib import contextmanager
from dataclasses import dataclass
from enum import StrEnum
from pathlib import Path
from tempfile import mkstemp
from typing import Callable, Generator

from reling.db.models import Language
from reling.scanner import Scanner
from reling.types import Input, Transcriber
from reling.utils.console import input_and_erase, interruptible_input, print_and_erase, print_and_maybe_erase
from reling.utils.prompts import ENTER, format_shortcut, Prompt
from .wave import FILE_EXTENSION, play, record

__all__ = [
    'get_input',
    'ScannerParams',
    'TranscriberParams',
]


class AudioAction(StrEnum):
    RE_RECORD = 're-record'
    LISTEN = 'listen'
    MANUAL_INPUT = 'manual input'


class ImageAction(StrEnum):
    RETAKE = 'retake'
    MANUAL_INPUT = 'manual input'


MANUAL_SHORTCUT = 'm'
assert MANUAL_SHORTCUT == AudioAction.MANUAL_INPUT.value[0] == ImageAction.MANUAL_INPUT.value[0]
SKIP_SHORTCUT = 's'

MULTIMEDIA_PROMPT = (f'({ENTER} to {{action}},'
                     f' {format_shortcut(MANUAL_SHORTCUT)} to input manually,'
                     f' {format_shortcut(SKIP_SHORTCUT)} to skip)')

START_RECORDING = 'start recording'
RECORDING_REDO_SHORTCUT = AudioAction.RE_RECORD.value[0]
RECORDING_UNTIL_ENTER = f'(recording... {ENTER} to stop, {format_shortcut(RECORDING_REDO_SHORTCUT)} to redo)'
TRANSCRIBING = '(transcribing...)'

CAPTURE_IMAGE = 'capture image'
CAPTURING = '(capturing...)'
PROCESSING = '(processing...)'

PAUSED = f'PAUSED ({ENTER} to resume)'


@dataclass
class TranscriberParams:
    transcribe: Transcriber
    storage: Path


@dataclass
class ScannerParams:
    scanner: Scanner
    language: Language


@dataclass
class ResultToReturn:
    result: Input | None


def prompt_multimedia(prefix: str, action: str) -> ResultToReturn | None:
    """Prompt the user for an action related to multimedia input."""
    response = input_and_erase(prefix + MULTIMEDIA_PROMPT.format(action=action)).strip().lower()
    if response == MANUAL_SHORTCUT.lower():
        return ResultToReturn(None)
    elif response == SKIP_SHORTCUT.lower():
        return ResultToReturn(Input(''))
    else:
        return None


@contextmanager
def pausing(on_pause: Callable[[], None], on_resume: Callable[[], None]) -> Generator[None, None, None]:
    """Pause and then resume once the context is exited."""
    on_pause()
    yield
    on_resume()


def get_manual_input(prompt: str, suggestions: list[str]) -> Input:
    """Prompt the user for manual input."""
    return Input(interruptible_input(prompt, suggestions, mouse_support=True))


def get_temp_file(storage: Path) -> Path:
    """Generate a temporary file in the given storage directory."""
    return Path(mkstemp(dir=storage, suffix=FILE_EXTENSION)[1])


def do_record(prompt: str, file: Path) -> bool:
    """Record audio and return whether the user is satisfied with the recording."""
    with record(file):
        return input_and_erase(prompt + RECORDING_UNTIL_ENTER).strip().lower() != RECORDING_REDO_SHORTCUT.lower()


def do_transcribe(prompt: str, transcribe: Transcriber, file: Path) -> str:
    """Transcribe the audio file and return the transcription."""
    with print_and_erase(prompt + TRANSCRIBING):
        try:
            return transcribe(file)
        except KeyboardInterrupt:
            return ''


def get_audio_input(
        prompt: str,
        params: TranscriberParams,
        on_pause: Callable[[], None],
        on_resume: Callable[[], None],
        dynamic_suggestions: list[str],
) -> Input | None:
    """
    Get input from the user via audio recording, with options for re-recording, listening, and manual input.
    :return: The user's input, or None if the user has chosen to manually input the text.
    """
    if result := prompt_multimedia(prompt, START_RECORDING):
        return result.result
    file = get_temp_file(params.storage)
    while True:
        if not do_record(prompt, file):
            continue
        with pausing(on_pause, on_resume):
            transcription = do_transcribe(prompt, params.transcribe, file)
            dynamic_suggestions.append(transcription)
        with print_and_maybe_erase(prompt + transcription) as do_erase:
            try:
                while True:
                    match Prompt.from_enum(AudioAction).prompt():
                        case AudioAction.RE_RECORD:
                            do_erase()
                            break
                        case AudioAction.LISTEN:
                            try:
                                play(file)
                            except KeyboardInterrupt:
                                pass
                        case AudioAction.MANUAL_INPUT:
                            do_erase()
                            return None
                        case None:
                            return Input(transcription, audio=file)
                        case _:
                            assert False
            except KeyboardInterrupt:
                do_erase()
                raise


def do_scan(prompt: str, scanner: Scanner, language: Language) -> str:
    """Capture an image and return the extracted text."""
    with print_and_erase(prompt + CAPTURING):
        try:
            image = scanner.capture()
        except KeyboardInterrupt:
            return ''
    with print_and_erase(prompt + PROCESSING):
        try:
            return scanner.process(image, language)
        except KeyboardInterrupt:
            return ''


def get_image_input(
        prompt: str,
        params: ScannerParams,
        on_pause: Callable[[], None],
        on_resume: Callable[[], None],
        dynamic_suggestions: list[str],
) -> Input | None:
    """
    Get input from the user via image capture, with options for retaking or manual input.
    :return: The user's input, or None if the user has chosen to manually input the text.
    """
    if result := prompt_multimedia(prompt, CAPTURE_IMAGE):
        return result.result
    while True:
        with pausing(on_pause, on_resume):
            text = do_scan(prompt, params.scanner, params.language)
            dynamic_suggestions.append(text)
        with print_and_maybe_erase(prompt + text) as do_erase:
            try:
                while True:
                    match Prompt.from_enum(ImageAction).prompt():
                        case ImageAction.RETAKE:
                            do_erase()
                            break
                        case ImageAction.MANUAL_INPUT:
                            do_erase()
                            return None
                        case None:
                            return Input(text)
                        case _:
                            assert False
            except KeyboardInterrupt:
                do_erase()
                raise


def get_input(
        on_pause: Callable[[], None],
        on_resume: Callable[[], None],
        prompt: str = '',
        input_suggestions: list[str] | None = None,
        transcriber_params: TranscriberParams | None = None,
        scanner_params: ScannerParams | None = None,
) -> Input:
    """Get user input, optionally via audio recording or image capture."""
    if transcriber_params and scanner_params:
        raise ValueError('Cannot use both transcriber and scanner.')
    dynamic_suggestions = [*(input_suggestions or [])]
    while True:
        try:
            return (
                (transcriber_params and get_audio_input(
                    prompt,
                    transcriber_params,
                    on_pause,
                    on_resume,
                    dynamic_suggestions,
                ))
                or
                (scanner_params and get_image_input(
                    prompt,
                    scanner_params,
                    on_pause,
                    on_resume,
                    dynamic_suggestions,
                ))
                or
                get_manual_input(prompt, dynamic_suggestions)
            )
        except KeyboardInterrupt:
            with pausing(on_pause, on_resume):
                input_and_erase(PAUSED)



__all__ = [
    'coalesce',
    'ensure_not_none',
    'wrap_in_list',
]


def ensure_not_none[T](value: T | None) -> T:
    """Raise a ValueError if the value is None."""
    if value is None:
        raise ValueError('Value cannot be None.')
    return value


def coalesce[T](*values: T | None) -> T | None:
    """Return the first non-null value, if any."""
    return next((value for value in values if value is not None), None)


def wrap_in_list[T](value: T | None) -> list[T]:
    """Wrap the value in a list if it is not None."""
    return [value] if value is not None else []



from dataclasses import dataclass

from reling.types import DialogueExchangeData, Input

__all__ = [
    'ExchangeWithTranslation',
    'ExplanationRequest',
    'PreScoreWithSuggestion',
    'ScoreWithSuggestion',
    'SentenceWithTranslation',
]


@dataclass
class SentenceWithTranslation:
    sentence: str
    translation: Input | None

    @property
    def input(self) -> Input | None:
        return self.translation

    def all(self) -> list[str]:
        return [self.sentence]

    def input_within_all(self) -> list[str | None]:
        return [self.translation.text if self.translation else None]


@dataclass
class ExchangeWithTranslation:
    exchange: DialogueExchangeData
    user_translation: Input | None

    @property
    def input(self) -> Input | None:
        return self.user_translation

    def all(self) -> list[str]:
        return list(self.exchange.all())

    def input_within_all(self) -> list[str | None]:
        self.exchange.assert_speaker_comes_first()
        return [None, self.user_translation.text if self.user_translation else None]


@dataclass
class PreScoreWithSuggestion:
    score: int
    suggestion: str | None


@dataclass
class ScoreWithSuggestion:
    score: int
    suggestion: str | None


@dataclass
class ExplanationRequest:
    sentence_index: int
    source: bool



from reling.db.models import Dialogue, Text
from reling.shelf import get_value, set_value

__all__ = [
    'get_default_content_id',
    'set_default_content',
]

VAR_NAME = 'DEFAULT_CONTENT'


def get_default_content_id() -> str | None:
    """Get the default content ID."""
    return get_value(VAR_NAME)


def set_default_content(content: Text | Dialogue | None) -> None:
    """Set the default content for the commands."""
    set_value(VAR_NAME, content.id if content is not None else None)



from typing import Literal

from openai import OpenAI

from reling.db.models import Language
from reling.helpers.openai import openai_handler
from reling.helpers.pyaudio import get_audio, get_stream
from reling.types import Speed
from .tts_client import TTSClient
from .voices import Voice

__all__ = [
    'OpenAITTSClient',
]

CHANNELS = 1
RATE = 24000
CHUNK_SIZE = 1024
RESPONSE_FORMAT: Literal['pcm'] = 'pcm'


class OpenAITTSClient(TTSClient):
    _client: OpenAI
    _model: str
    _language: Language

    def __init__(self, *, api_key: str, model: str, language: Language) -> None:
        self._client = OpenAI(api_key=api_key)
        self._model = model
        self._language = language

    def read(self, text: str, voice: Voice, speed: Speed) -> None:
        with (
            get_audio() as pyaudio,
            get_stream(
                pyaudio=pyaudio,
                format=pyaudio.paInt16,
                channels=CHANNELS,
                rate=RATE,
                output=True,
            ) as stream,
            openai_handler(),
        ):
            response = self._client.audio.speech.create(
                model=self._model,
                voice=voice.value,  # type: ignore
                response_format=RESPONSE_FORMAT,
                input=text,
                speed=speed.value,
                instructions=f'Read in {self._language.name}.',
            )
            stream.write(response.content)



from datetime import datetime
import re
from typing import cast, Generator

from sqlalchemy import ColumnElement

from reling.app.app import app
from reling.app.types import (
    ARCHIVE_OPT,
    CONTENT_CATEGORY_OPT,
    IDS_ONLY_OPT,
    LANGUAGE_OPT,
    LEVEL_OPT,
    REGEX_CONTENT_OPT,
)
from reling.db import Session, single_session
from reling.db.enums import ContentCategory, Level
from reling.db.models import Dialogue, Language, Text
from reling.utils.time import format_time
from reling.utils.tables import build_table, print_table

__all__ = [
    'list_',
]

ID = 'ID'
LANGUAGE = 'Language'
LEVEL = 'Level'
TOPIC = 'Topic'
STYLE = 'Style'
SPEAKER = 'Speaker'
SIZE = 'Size'
CREATED_AT = 'Created at'
ARCHIVED_AT = 'Archived at'

NO_TOPIC = 'N/A'


def get_text(item: Text | Dialogue) -> list[str]:
    """Return the text content of a text or dialogue, including the speaker, topic, style, and sentences."""
    if isinstance(item, Text):
        return [
            item.topic,
            item.style,
            *(sentence.sentence for sentence in item.sentences),
        ]
    else:
        return [
            item.speaker,
            *([item.topic] if item.topic is not None else []),
            *(turn for exchange in item.exchanges for turn in [exchange.speaker, exchange.user]),
        ]


def match(item: Text | Dialogue, level: Level | None, language: Language | None, search: re.Pattern | None) -> bool:
    return (
        (level is None or item.level == level)
        and (language is None or cast(Language, item.language).id == language.id)
        and (search is None
             or search.search(cast(str, item.id)) is not None
             or any(search.search(text) is not None for text in get_text(item)))
    )


def find_items[T: type[Text | Dialogue]](
        session: Session,
        model: type[T],
        archive: bool,
        level: Level | None,
        language: Language | None,
        search: re.Pattern | None,
) -> Generator[T, None, None]:
    archived_at = cast(ColumnElement[datetime | None], model.archived_at)
    created_at = cast(ColumnElement[datetime], model.created_at)

    for item in session.query(model).filter(
        archived_at.is_not(None) if archive else archived_at.is_(None),
    ).order_by(
        archived_at.desc(), created_at.desc(),
    ):
        if match(item, level, language, search):
            yield item


@app.command(name='list')
def list_(
        category: CONTENT_CATEGORY_OPT = None,
        level: LEVEL_OPT = None,
        language: LANGUAGE_OPT = None,
        search: REGEX_CONTENT_OPT = None,
        archive: ARCHIVE_OPT = False,
        ids_only: IDS_ONLY_OPT = False,
) -> None:
    """List texts and/or dialogues, optionally filtered by ID or other criteria."""
    with single_session() as session:
        for model in [
            *([Text] if category != ContentCategory.DIALOGUE else []),
            *([Dialogue] if category != ContentCategory.TEXT else []),
        ]:
            items = find_items(session, model, archive, level, language, search)
            if ids_only:
                for item in items:
                    print(item.id)
            else:
                table = build_table(
                    title=('Texts' if model is Text else 'Dialogues') + (' (archived)' if archive else ''),
                    headers=[
                        ID,
                        LANGUAGE,
                        LEVEL,
                        *([SPEAKER] if model is Dialogue else []),
                        TOPIC,
                        *([STYLE] if model is Text else []),
                        SIZE,
                        CREATED_AT,
                        *([ARCHIVED_AT] if archive else []),
                    ],
                    justify={
                        SIZE: 'right',
                    },
                    data=({
                        ID: item.id,
                        LANGUAGE: item.language.name,
                        LEVEL: item.level.value,
                        **({SPEAKER: item.speaker} if model is Dialogue else {}),
                        TOPIC: item.topic or NO_TOPIC,
                        **({STYLE: item.style} if model is Text else {}),
                        SIZE: str(len(item.sentences if model is Text else item.exchanges)),
                        CREATED_AT: format_time(item.created_at),
                        **({ARCHIVED_AT: format_time(item.archived_at)} if archive else {}),
                    } for item in items),
                )
                print()
                print_table(table)



from pathlib import Path
from typing import Generator

__all__ = [
    'read_csv',
]

DEFAULT_DELIMITER = ','


def read_csv(
        path: Path,
        columns: list[str],
        *,
        empty_as_none: bool = False,
        delimiter: str = DEFAULT_DELIMITER,
) -> Generator[dict[str, str | None], None, None]:
    """Read a CSV file and yield dictionaries with column names as keys."""
    with path.open('r') as file:
        for line in file:
            yield {
                column: value if value != '' or not empty_as_none else None
                for column, value in zip(columns, line.strip('\n').split(delimiter))
            }



from typing import Callable, Generator

from reling.config import MAX_SCORE
from reling.db.enums import ContentCategory
from reling.db.models import Language
from reling.gpt import GPTClient
from reling.types import DialogueExchangeData, Promise
from reling.utils.feeders import CharFeeder
from reling.utils.values import coalesce
from .types import ExchangeWithTranslation, ExplanationRequest, ScoreWithSuggestion, SentenceWithTranslation

__all__ = [
    'build_explainer',
]

MISTAKE_THRESHOLD_RATIO = 0.5


def if_not(text: str | None, until_equals: str) -> str | None:
    """Return the text if it does not equal the specified value, otherwise return None."""
    return text if text != until_equals else None


def explain_structure(sentence: str, language: Language) -> list[str]:
    """Generate a prompt section for explaining the structure of a translation in the specified language."""
    return [
        f'The last line is translated into {language.name} as follows:',
        f'"""{sentence}"""',
        f'Explain the structure of this translation to a learner of {language.name}.',
    ]


def explain_mistakes(sentence: str, corrected: str, language: Language, source_language: Language) -> list[str]:
    """Generate a prompt section explaining why a corrected translation is preferred over the provided one."""
    return [
        f'A learner of {language.name} translated the last line as follows:',
        f'"""{sentence}"""',
        f'',
        f'Compare it to the following reference translation:',
        f'"""{corrected}"""',
        f'',
        f'Enumerate the individual differences between the two versions. For each difference:',
        f'- Determine whether it is a mistake.',
        f'- If it is a mistake, specify its type:',
        f'  • **Grammatical error** – the learner’s translation is ungrammatical.',
        f'  • **Unidiomatic phrasing** – the translation is unnatural or awkward.',
        f'  • **Deviation from the original meaning** – the translation alters the intended message.',
        f'- If the mistake is a deviation from the original meaning, translate the learner’s version '
        f'(or relevant parts of it) back into {source_language.name} to clarify the difference.',
    ]


def explain_difference(sentence: str, alternative: str, language: Language) -> list[str]:
    """Generate a prompt section for explaining differences between a provided and an alternative translation."""
    return [
        f'A learner of {language.name} translated the last line as follows:',
        f'"""{sentence}"""',
        f'',
        f'Compare it to the following alternative translation:',
        f'"""{alternative}"""',
        f'',
        f'Enumerate the individual differences between the two versions. For each difference:',
        f'- Explain whether it affects meaning, grammar, or naturalness.',
        f'- Indicate which version is more appropriate and why.',
        f'- If both versions are acceptable, clarify the nuances in their usage and context.',
    ]


def explain_source_structure(sentence: str) -> list[str]:
    """Generate a prompt section for explaining the structure of a sentence in the source language."""
    return [
        f'Explain the grammatical and structural elements of the last line ("""{sentence}""") to a language learner.',
    ]


def do_explain(
        category: ContentCategory,
        gpt: Promise[GPTClient],
        initial_sentences: list[str],
        provided: str,
        original: str,
        result: ScoreWithSuggestion,
        source_language: Language,
        target_language: Language,
        explain_source: bool,
) -> Generator[str, None, None]:
    """Generate an explanation for translations based on user-provided and corrected data."""
    return gpt().ask(
        prompt='\n'.join([
            f'Here is the beginning of a {category.value} in {source_language.name}:',
            f'',
            *initial_sentences,
            f'',
            *(
                explain_source_structure(initial_sentences[-1])
                if explain_source else
                explain_difference(
                    sentence=provided,
                    alternative=alternative,
                    language=target_language,
                )
                if result.score == MAX_SCORE and (alternative := coalesce(
                    if_not(result.suggestion, provided),
                    if_not(original, provided),
                )) else
                explain_mistakes(
                    sentence=provided,
                    corrected=result.suggestion or original,
                    language=target_language,
                    source_language=source_language,
                )
                if MISTAKE_THRESHOLD_RATIO * MAX_SCORE < result.score < MAX_SCORE else
                explain_structure(
                    sentence=result.suggestion or original,
                    language=target_language,
                )
            ),
        ]),
        creative=False,
        feeder_type=CharFeeder,
        auto_normalize=False,
    )


def build_explainer(
        category: ContentCategory,
        gpt: Promise[GPTClient],
        items: list[SentenceWithTranslation | ExchangeWithTranslation],
        original_translations: list[str | DialogueExchangeData],
        results: list[ScoreWithSuggestion | None],
        source_language: Language,
        target_language: Language,
) -> Callable[[ExplanationRequest], Generator[str, None, None]]:
    """Create a function to generate explanations for text or dialogue translations."""
    is_text = category == ContentCategory.TEXT

    def explain(request: ExplanationRequest) -> Generator[str, None, None]:
        index = request.sentence_index
        assert results[index] is not None
        return do_explain(
            category=category,
            gpt=gpt,
            initial_sentences=(
                [item.sentence for item in items[:index + 1]]
                if is_text
                else [turn for item in items[:index + 1] for turn in item.exchange.all()]
            ),
            provided=items[index].input.text,
            original=original_translations[index] if is_text else original_translations[index].user,
            result=results[index],
            source_language=source_language,
            target_language=target_language,
            explain_source=request.source,
        )

    return explain



from typing import Generator

from openai import OpenAI

from reling.gpt_log import GptLogItem, log
from reling.helpers.openai import openai_handler
from reling.types import Image
from reling.utils.feeders import Feeder, LineFeeder
from reling.utils.transformers import normalize, Transformer

__all__ = [
    'GPTClient',
]

CREATIVE_TEMPERATURE = 1.0
INCOMPLETE_LOG_MARKER = '\n...'


class GPTClient:
    _client: OpenAI
    _model: str

    def __init__(self, *, api_key: str, model: str) -> None:
        self._client = OpenAI(api_key=api_key)
        self._model = model

    def ask(
            self,
            prompt: str,
            image: Image | None = None,
            creative: bool = True,
            feeder_type: type[Feeder] = LineFeeder,
            transformers: list[Transformer] | None = None,
            auto_normalize: bool = True,
    ) -> Generator[str, None, None]:
        """
        Ask the model a question and yield sections of the response as they become available, applying transformers.
        """
        feeder = feeder_type()
        temperature = CREATIVE_TEMPERATURE if creative else 0.0

        with openai_handler():
            stream = self._client.chat.completions.create(
                model=self._model,
                stream=True,
                messages=[{'role': 'user', 'content': [
                    {'type': 'text', 'text': prompt},
                    *([{'type': 'image_url', 'image_url': {'url': image.get_url()}}] if image else []),
                ]}],
                temperature=temperature,
            )

        section_index = 0

        def flush() -> Generator[str, None, None]:
            nonlocal section_index
            while (section := feeder.get()) is not None:
                for transformer in (transformers or []) + ([normalize] if auto_normalize else []):
                    section = transformer(section, section_index)
                    if section is None:
                        break
                else:
                    yield section
                    section_index += 1

        chunks: list[str] = []
        complete = False
        try:
            for chunk in stream:
                content = chunk.choices[0].delta.content or ''
                chunks.append(content)
                feeder.put(content)
                yield from flush()

            feeder.end()
            yield from flush()
            complete = True
        finally:
            log(GptLogItem(
                prompt=prompt,
                temperature=temperature,
                response=''.join(chunks + ([INCOMPLETE_LOG_MARKER] if not complete else [])),
            ))



from sqlalchemy import and_, ColumnElement

from reling.db.models import DialogueExam, Language, TextExam
from .modalities import Modality

__all__ = [
    'get_filter',
]


def get_filter(
        language: Language,
        paired: list[Language] | None,
        modality: Modality,
        model: type[TextExam | DialogueExam],
) -> ColumnElement[bool]:
    """Get the filtering condition for the given language, paired language(s), modality, and model."""
    main_language_id, secondary_language_id = ((model.source_language_id, model.target_language_id)
                                               if modality == Modality.COMPREHENSION else
                                               (model.target_language_id, model.source_language_id))
    return and_(main_language_id == language.id,
                paired is None or secondary_language_id.in_(language.id for language in paired))



from reling.app.app import app
from reling.app.default_content import set_default_content
from reling.app.types import (
    API_KEY,
    ASR_MODEL,
    EXAM_CONTENT_ARG,
    ExamExtraContentOptions,
    HIDE_PROMPTS_OPT,
    LANGUAGE_OPT,
    LANGUAGE_OPT_FROM,
    LISTEN_OPT,
    MODEL,
    OFFLINE_SCORING_OPT,
    READ_LANGUAGE_OPT,
    RETRY_OPT,
    SCAN_OPT,
    SKIP_OPT,
    TTS_MODEL,
)
from reling.asr import ASRClient
from reling.db.models import Dialogue, Language, Text
from reling.gpt import GPTClient
from reling.helpers.audio import ensure_audio
from reling.helpers.typer import typer_raise
from reling.scanner import ScannerManager, ScannerParams
from reling.tts import get_tts_client
from reling.utils.prompts import enter_to_continue
from .execution import perform_exam
from .repetition import compute_repetition_data, print_repetition_statistics, RepetitionData
from .skips import get_skipped_indices

__all__ = [
    'exam',
]


def adjust_exam_params(
        content: Text | Dialogue | ExamExtraContentOptions,
        from_: Language | None,
        to: Language | None,
        skip: int | None,
        read: list[Language] | None,
        listen: bool,
        scan: int | None,
) -> tuple[
        tuple[
            Text | Dialogue,
            Language,
            Language,
            set[int],
            list[Language],
            bool,
            int | None,
        ] | None,
        RepetitionData | None,
]:
    """Validate and adjust exam parameters; handle spaced repetition mode if requested."""
    read = read or []
    if listen and scan is not None:
        typer_raise('Choose either listen or scan, not both.')
    if read or listen:
        ensure_audio()

    if content == ExamExtraContentOptions.SPACED_REPETITION:
        if skip is not None:
            typer_raise('Cannot skip sentences in spaced repetition mode.')
        repetition_data = compute_repetition_data(source_language=from_, target_language=to)
        if next_exam := repetition_data.next_exam:
            content = next_exam.content
            from_ = next_exam.source_language
            to = next_exam.target_language
            skipped_indices = next_exam.skipped_indices
        else:
            return None, repetition_data
    else:
        if from_ is None and to is None:
            typer_raise('You must specify at least one language.')
        from_ = from_ or content.language
        to = to or content.language
        if from_ == to:
            typer_raise('Source and target languages must be different.')
        repetition_data = None
        skipped_indices = (get_skipped_indices(content, source_language=from_, target_language=to, skip_after=skip)
                           if skip is not None else set())

    for language in read:
        if language not in [from_, to]:
            typer_raise(f'Cannot read in {language.name} as the exam is from {from_.name} to {to.name}.')

    return (content, from_, to, skipped_indices, read, listen, scan), repetition_data


@app.command()
def exam(
        api_key: API_KEY,
        model: MODEL,
        tts_model: TTS_MODEL,
        asr_model: ASR_MODEL,
        content: EXAM_CONTENT_ARG,
        from_: LANGUAGE_OPT_FROM = None,
        to: LANGUAGE_OPT = None,
        skip: SKIP_OPT = None,
        read: READ_LANGUAGE_OPT = None,
        listen: LISTEN_OPT = False,
        scan: SCAN_OPT = None,
        hide_prompts: HIDE_PROMPTS_OPT = False,
        offline_scoring: OFFLINE_SCORING_OPT = False,
        retry: RETRY_OPT = False,
) -> None:
    """
    Test the user's ability to translate content from one language to another.
    If only one language is specified, the content's original language is assumed for the unspecified direction.
    """
    params, repetition_data = adjust_exam_params(content, from_, to, skip, read, listen, scan)

    if repetition_data:
        print_repetition_statistics(repetition_data.statistics)
        print()

    if not params:
        typer_raise('No content to review, exiting.', is_error=False)

    content, from_, to, skipped_indices, read, listen, scan = params

    if repetition_data:
        print(f'Continuing with "{content.id}"...')
        print()
        enter_to_continue()

    set_default_content(content)

    if len(skipped_indices) == content.size:
        typer_raise('All sentences are skipped, exiting.', is_error=False)

    def get_gpt() -> GPTClient:
        return GPTClient(api_key=api_key.get(), model=model.get())

    perform_exam(
        get_gpt,
        content,
        skipped_indices=skipped_indices,
        source_language=from_,
        target_language=to,
        source_tts=(get_tts_client(model=tts_model.get(), api_key=api_key.promise(), language=from_)
                    if from_ in read else None),
        target_tts=(get_tts_client(model=tts_model.get(), api_key=api_key.promise(), language=to)
                    if to in read else None),
        asr=ASRClient(api_key=api_key.get(), model=asr_model.get()) if listen else None,
        scanner_manager=ScannerManager(ScannerParams(
            camera_index=scan,
            gpt=get_gpt(),
        ) if scan is not None else None),
        hide_prompts=hide_prompts,
        offline_scoring=offline_scoring,
        retry=retry,
    )



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import cast

from reling.app.exceptions import AlgorithmException
from reling.app.translation import get_dialogue_exchanges, get_text_sentences
from reling.asr import ASRClient
from reling.config import MAX_SCORE
from reling.db.enums import ContentCategory, Gender
from reling.db.models import Dialogue, DialogueExam, Language, Text, TextExam
from reling.gpt import GPTClient
from reling.helpers.typer import typer_raise
from reling.helpers.voices import pick_voices
from reling.scanner import Scanner, ScannerManager
from reling.tts import TTSClient, TTSVoiceClient
from reling.types import DialogueExchangeData, Promise
from reling.utils.timetracker import TimeTracker
from .explanation import build_explainer
from .input import collect_translations
from .presentation import present_results
from .scoring import score_translations
from .storage import save_exam
from .types import ExchangeWithTranslation, ScoreWithSuggestion, SentenceWithTranslation

__all__ = [
    'perform_exam',
]


def collect_perfect(content: Text | Dialogue, target_language: Language) -> list[set[str]]:
    """
    Collect the suggestions and correct answers from previous exams in the same target language, indexed by sentence.
    """
    suggestions = [set() for _ in range(content.size)]
    for exam in cast(list[TextExam] | list[DialogueExam], content.exams):
        if exam.target_language == target_language:
            for result in exam.results:
                if result.suggested_answer:
                    suggestions[result.index].add(result.suggested_answer)
                if result.score == MAX_SCORE:
                    suggestions[result.index].add(result.answer)
    return suggestions


def get_voices(
        content: Text | Dialogue,
        source_tts: TTSClient | None,
        target_tts: TTSClient | None,
) -> tuple[TTSVoiceClient | None, TTSVoiceClient | None, TTSVoiceClient | None]:
    """
    Pick TTS voices based on the content type (text or dialogue):
    - For a Text, return (source_voice, target_voice, None).
    - For a Dialogue, return (source_user_voice, target_user_voice, target_speaker_voice).
    """
    if isinstance(content, Text):
        source_voice, target_voice = pick_voices(None, None)
        return (
            source_tts.with_voice(source_voice) if source_tts else None,
            target_tts.with_voice(target_voice) if target_tts else None,
            None,
        )
    else:
        speaker_voice, user_voice = pick_voices(
            cast(Gender, content.speaker_gender),
            cast(Gender, content.user_gender),
        )
        return (
            source_tts.with_voice(user_voice) if source_tts else None,
            target_tts.with_voice(user_voice) if target_tts else None,
            target_tts.with_voice(speaker_voice) if target_tts else None,
        )


def perform_exam_round(
        gpt: Promise[GPTClient],
        content: Text | Dialogue,
        items: list[str | DialogueExchangeData],
        original_translations: list[str | DialogueExchangeData],
        skipped_indices: set[int],
        source_language: Language,
        target_language: Language,
        source_tts: TTSVoiceClient | None,
        target_speaker_tts: TTSVoiceClient | None,
        asr: ASRClient | None,
        scanner_manager: ScannerManager,
        hide_prompts: bool,
        offline_scoring: bool,
        previous_attempts: list[list[SentenceWithTranslation | ExchangeWithTranslation]],
        previous_scores: list[list[ScoreWithSuggestion]],
        storage: Path,
        tracker: TimeTracker,
) -> tuple[
    list[SentenceWithTranslation | ExchangeWithTranslation],
    list[ScoreWithSuggestion | None],
    Scanner | None,
]:
    """Collect user translations of the text or dialogue, score them, and return the results, all in a single round."""
    with scanner_manager.get_scanner() as scanner:
        tracker.resume()
        translated = list(collect_translations(
            category=ContentCategory.TEXT if isinstance(content, Text) else ContentCategory.DIALOGUE,
            items=items,
            original_translations=original_translations,
            skipped_indices=skipped_indices,
            target_language=target_language,
            source_tts=source_tts,
            target_speaker_tts=target_speaker_tts,
            asr=asr,
            scanner=scanner,
            hide_prompts=hide_prompts,
            previous_attempts=previous_attempts,
            previous_scores=previous_scores,
            storage=storage,
            on_pause=tracker.pause,
            on_resume=tracker.resume,
        ))
        tracker.pause()

    try:
        results = list(score_translations(
            category=ContentCategory.TEXT if isinstance(content, Text) else ContentCategory.DIALOGUE,
            gpt=gpt,
            items=translated,
            original_translations=original_translations,
            previous_perfect=collect_perfect(content, target_language),
            source_language=source_language,
            target_language=target_language,
            offline=offline_scoring,
        ))
    except AlgorithmException as e:
        typer_raise(e.msg)

    return translated, results, scanner


def perform_exam(
        gpt: Promise[GPTClient],
        content: Text | Dialogue,
        skipped_indices: set[int],
        source_language: Language,
        target_language: Language,
        source_tts: TTSClient | None,
        target_tts: TTSClient | None,
        asr: ASRClient | None,
        scanner_manager: ScannerManager,
        hide_prompts: bool,
        offline_scoring: bool,
        retry: bool,
) -> None:
    """
    Collect user translations of the text or dialogue, score them, save and present the results to the user,
    optionally reading the source and/or target language out loud.
    """
    with TemporaryDirectory() as file_storage:
        is_text = isinstance(content, Text)

        voice_source_tts, voice_target_tts, voice_target_speaker_tts = get_voices(content, source_tts, target_tts)
        items, original_translations = (
            (get_text_sentences if is_text else get_dialogue_exchanges)(content, language, gpt)
            for language in (source_language, target_language)
        )

        updated_skipped_indices = {*skipped_indices}

        translated: list[SentenceWithTranslation | ExchangeWithTranslation] | None = None
        results: list[ScoreWithSuggestion] | None = None

        previous_attempts: list[list[SentenceWithTranslation | ExchangeWithTranslation]] = []
        previous_scores: list[list[ScoreWithSuggestion]] = []

        tracker = TimeTracker()
        tracker.pause()
        while True:
            current_translated, current_results, scanner = perform_exam_round(
                gpt=gpt,
                content=content,
                items=items,
                original_translations=original_translations,
                skipped_indices=updated_skipped_indices,
                source_language=source_language,
                target_language=target_language,
                source_tts=voice_source_tts,
                target_speaker_tts=voice_target_speaker_tts,
                asr=asr,
                scanner_manager=scanner_manager,
                hide_prompts=hide_prompts,
                offline_scoring=offline_scoring,
                previous_attempts=previous_attempts,
                previous_scores=previous_scores,
                storage=Path(file_storage),
                tracker=tracker,
            )

            if translated is None:
                translated = [*current_translated]
                results = [*current_results]
            else:
                for index, (attempt, score) in enumerate(zip(current_translated, current_results)):
                    if attempt.input and attempt.input.text and score and score.score >= results[index].score:
                        translated[index] = attempt
                        results[index] = score

            for index, (attempt, score) in enumerate(zip(current_translated, current_results)):
                if (score and score.score == MAX_SCORE) or (attempt.input and attempt.input.text == ''):
                    updated_skipped_indices.add(index)

            if not retry or len(updated_skipped_indices) == content.size:
                break

            previous_attempts.append(current_translated)
            previous_scores.append(current_results)
        tracker.stop()

        exam = save_exam(
            content=content,
            source_language=source_language,
            target_language=target_language,
            read_source=source_tts is not None,
            read_target=target_tts is not None,
            listened=asr is not None,
            scanned=scanner is not None,
            started_at=tracker.started_at,
            finished_at=tracker.finished_at,
            total_pause_time=tracker.total_pause_time,
            items=translated,
            results=results,
        )

        present_results(
            items=translated,
            original_translations=original_translations,
            exam=exam,
            source_tts=voice_source_tts,
            target_tts=voice_target_tts,
            target_speaker_tts=voice_target_speaker_tts,
            explain=build_explainer(
                category=ContentCategory.TEXT if is_text else ContentCategory.DIALOGUE,
                gpt=gpt,
                items=translated,
                original_translations=original_translations,
                results=results,
                source_language=source_language,
                target_language=target_language,
            ),
        )
